# -*- coding: iso-8859-1 -*-
# TODO: wg. robustheit mit logs rechnen, vgl. Programmierprfung SS07
"""
Ein Programm zum Rumspielen mit Mixtures von Gaussians mit diagonaler
Kovarianzmatrix.

Vgl. Skript, "Exkurs: Expectation Maximization"
"""

import operator, math, bisect, random
from itertools import *

def normalize(seq):
	"""changes seq in place such that the sum of its elements is one.
	"""
	total = float(sum(seq))
	if total==0:
		raise ValueError, "Un-normalizable sequence"
	for i in xrange(len(seq)):
		seq[i] = seq[i]/total


def computeWeightedMean(seq, weights, weightSum=None):
	"""returns a mean of the elements of seq weighted by weights (i.e.
	sum(seq[i]*weights[i])/weightSum).  You can pass weightSum to
	save the time for recomputing the sum of all weights.
	"""
	if weightSum is None:
		weightSum = sum(weights)
	return reduce(operator.add, starmap(operator.mul, izip(seq, weights))
		)/weightSum


class Point(tuple):
	"""This class encapsulates a Point in an n-Dimensional space.  The
	basic idea is that some of python's operators do the stuff we want
	them to do on vectors.  We inherit the constructor from tuple, so
	you want to instantiate from a sequence.
	"""
	def __mul__(self, other):
		"""returns the (other*self[i]) if other is a scalar (i.e. something
		we can't iterate over, sum(other[i]*self[i]) (the standard scalar 
		product) otherwise.
		"""
		if hasattr(other, "__iter__"):
			# scalar product
			return sum(starmap(operator.mul, izip(self, other)))
		else:
			return Point(imap(lambda a:other*a, self))

	def __div__(self, other):
		"""returns the result of a component-wise division.
		"""
		return Point(imap(lambda a:a/other, self))

	def __add__(self, other):
		"""returns the vector sum, i.e. other has to be iterable.
		"""
		return Point(starmap(operator.add, izip(self, other)))

	def __sub__(self, other):
		"""returns the vector difference, i.e. other has to be iterable.
		"""
		return Point(starmap(operator.sub, izip(self, other)))

	def __pow__(self, other):
		"""returns the component-wise power, i.e. other has to be a
		scalar.
		"""
		return Point(imap(lambda a:a**other, self))


class DistFunc:
	"""This class represents a distribution function generated from
	a distribution.  Its draw method will return items distributed
	according to this distribution.
	"""
	def __init__(self, dist):
		total = float(sum(dist))
		self.dist = [v/total for v in dist]
		self.df = []
		last = 0
		for index, val in enumerate(self.dist):
			if val>1e-10:
				last += val
				self.df.append((last, index))

	def draw(self):
		return self.df[bisect.bisect(
				self.df, (random.random(),0))][1]

	def sampleRandomly(self, samplen):
		for i in xrange(samplen):
			yield self.draw()
	
	def getDist(self):
		"""returns the distribution for the distribution function.
		"""
		return self.dist


class Gaussian:
	"""This class represents one n-dimensional Gaussian.  To keep the
	number of parameters down, we assume the covariance matrix to be
	diagonal (i.e. the main axes are parallel to the coordinate axes).

	center and sigmas has to be a Point, sigmas a sequence of the same 
	length, viz., the number of dimensions of the space our data points 
	live in.

	To guard against degeneration, you can pass a minSigma gives
	the smallest value any sigmas component may have.
	"""
	def __init__(self, center, sigmas, minSigma=0):
		self.minSigma = minSigma
		self.setCenter(center)
		self.setSigmas(sigmas)

	def __str__(self):
		return "%s\n  %s"%(self.getCenter(), self.getSigmas())

	def __call__(self, point):
		"""returns the value of the pdf for point under this Gaussian.
		"""
		par = ((point-self.center)**2)*self.invDiagonal
		return math.exp(-0.5*par)/self.normalizer

	def setCenter(self, center):
		"""sets a new center for this gaussian.
		"""
		self.center = center

	def getCenter(self):
		"""returns the center of this gaussian.
		"""
		return self.center

	def setSigmas(self, sigmas):
		"""sets the diagonal of the covariance matrix.  sigmas has to
		be a sequence of proper length.
		"""
		self.sigmas = tuple([float(max(s, self.minSigma)) for s in sigmas])
		self.invDiagonal = [1/s**2 for s in self.sigmas]
		self.sigDet = reduce(operator.mul, self.sigmas)
		self.normalizer = math.sqrt((2*math.pi)**len(sigmas)*self.sigDet)

	def getSigmas(self):
		"""returns the individual sigmas.
		"""
		return self.sigmas

	def draw(self):
		"""returns one point drawn from this Gaussian distribution.
		"""
		return Point([random.gauss(mu, sigma) 
			for mu, sigma in zip(self.center, self.sigmas)])

	def update(self, pointSet, membershipProbs):
		weightSum = sum(membershipProbs)
		self.setCenter(computeWeightedMean(pointSet, membershipProbs, 
			weightSum))
		self.setSigmas(computeWeightedMean([(point-self.center)**2 
			for point in pointSet], membershipProbs, weightSum)**0.5)


class DistMixture:
	"""This class embeds a list of Distributions and their priors and knows
	how to handle them to perform an EM.
	"""
	def __init__(self, updatePriors=True):
		self.updatePriors = updatePriors
		self.distris = []
		self.priors = []
		for attName in ["__getitem__", "__len__"]:
			setattr(self, attName, getattr(self.distris, attName))

	def __str__(self):
		return "\n".join(["%f %s"%(prior, str(dist)) 
			for dist, prior in self])

	def __iter__(self):
		"""iterating over a DistMixture returns tuples of Distributions and
		the associated prior.
		"""
		return izip(self.distris, self.priors)

	def addDistribution(self, distri):
		"""adds a distribution to the mixture.  After an addition, the prior
		is a uniform distribution.
		"""
		self.distris.append(distri)
		self.priors = [1/float(len(self.distris))]*len(self.distris)

	def removeDistribution(self, distri):
		"""removes a distribution from the mixture.  The prior is simply
		renormalized.
		"""
		delInd = self.distris.index(distris)
		del self.priors[delInd]
		normalize(self.priors)
		del self.distris[delInd]

	def generatePointSet(self, nPoints):
		"""returns a list of Points that are distributed according
		to this DistributionMixture (only works if all underlying
		Distributions know how to draw).
		"""
		priorDist = DistFunc(self.priors)
		res = []
		for i in xrange(nPoints):
			res.append(self.distris[priorDist.draw()].draw())
		return res

	def _runEStep(self, pointSet):
		"""runs the expectation step and returns a list of lists containing
		the individual membership probabilities.  The index to the embedding
		list is the index of the distribution, the index to the embedded lists
		is the index of the point in pointSet.
		"""
		res = []
		for point in pointSet:
			probsForThisPoint = []
			for g, p in self:
				probsForThisPoint.append(g(point))
			normalize(probsForThisPoint)
			res.append(probsForThisPoint)
		return res

	def _runMStep(self, membershipProbs, pointSet):
		"""runs the maximization step by calling the update method
		of all distributions with the membership likelihoods of the
		points.
		"""
		for distInd, distri in enumerate(self.distris):
			probsForThisDist = [pfp[distInd] for pfp in membershipProbs]
			distri.update(pointSet, probsForThisDist)
			if self.updatePriors:
				self.priors[distInd] = sum(probsForThisDist)/float(len(pointSet))

	def runOneEmStep(self, pointSet):
		"""calls _runEStep and _runMStep for pointSet, thus updating
		the mixture for this set.
		"""
		membershipProbs = self._runEStep(pointSet)
		self._runMStep(membershipProbs, pointSet)

	def getTotalLogL(self, pointSet):
		"""returns the log of the total probability of pointSet -- this is the
		quantity that EM should maximize.
		"""
		return sum(imap(lambda point: 
			math.log(max(1e-207, sum(
					starmap(lambda distri, prior:prior*distri(point), self)
				))), 
				pointSet))

	def getMostLikelyCluster(self, point):
		"""returns the index of the distribution that has the highest
		likelihood for point.
		"""
		probs = [(distri(point), index) 
			for index, distri in enumerate(self.distris)]
		probs.sort()
		return probs[-1][1]


import Tkinter

class PointDisplay(Tkinter.Frame):
	"""Boring GUI stuff.  Basically: Ignore.
	"""
	def __init__(self, master, scale, *args, **kwargs):
		Tkinter.Frame.__init__(self, master, *args, **kwargs)
		self.scale = scale
		self.canvas = Tkinter.Canvas(self, background="#ffffff")
		self.canvas.pack(fill=Tkinter.BOTH, expand=1)
		self.points = []
		self.ells = []
		self.pointColor = "#000000"
		self.canvas.bind("<ButtonPress-1>", self._startdrag)
		self.canvas.bind("<B1-Motion>", self._duringdrag)
		self.canvas.bind("<ButtonRelease-1>", master._enterNewGaussian)
		self.canvas.bind("<ButtonPress-2>", master._setNewPoint)
	
	def _startdrag(self, ev):
		self._startx, self._starty = ev.x, ev.y
		self._draggedEllipse = self.canvas.create_oval(self._startx, 
			self._starty, self._startx+1, self._starty+1, 
			fill="#cccccc", outline="#cccccc")
	
	def _duringdrag(self, ev):
		if hasattr(self, "_draggedEllipse"):
			self.canvas.coords(self._draggedEllipse, self._startx, 
				self._starty, ev.x, ev.y)

	def getDraggedEllipse(self, ev):
		res = None
		if hasattr(self, "_draggedEllipse"):
			res = [c/float(self.scale) for c in [self._startx, self._starty,
				ev.x, ev.y]]
			self.canvas.delete(self._draggedEllipse)
			for attName in ["_startx", "_starty", "_draggedEllipse"]:
				delattr(self, attName)
		return res

	def getCoos(self, ev):
		return [c/float(self.scale) for c in [ev.x, ev.y]]

	def setPoint(self, x, y, color=None):
		if color is None:
			color = self.pointColor
		x, y = int(self.scale*x), int(self.scale*y)
		self.points.append(self.canvas.create_rectangle(
			[x-1, y-1, x+1, y+1], fill=color, outline=color))

	def clearPoints(self):
		for p in self.points:
			self.canvas.delete(p)
		self.points = []

	def setEllipse(self, x0, y0, x1, y1, color, callback):
		x0, y0, x1, y1 = [c*self.scale for c in [x0, y0, x1, y1]]
		self.ells.append(self.canvas.create_oval(
			x0, y0, x1, y1, fill=None, outline=color, width=2))
		self.canvas.tag_bind(self.ells[-1], "<Button-3>", callback)

	def clearEllipses(self):
		for e in self.ells:
			self.canvas.delete(e)
		self.ells = []


class TwoDMixtureVisualizer(Tkinter.Tk):
	"""Boring GUI stuff, and not too beautiful, either.  Ignore.
	"""
	def __init__(self, width, height, scale, *args, **kwargs):
		Tkinter.Tk.__init__(self, *args, **kwargs)
		self.doColor = 0
		self.geometry("%dx%d"%(width, height))
		self.pointDisplay = PointDisplay(self, scale)
		self.pointDisplay.grid(row=0, sticky=Tkinter.NE+Tkinter.SW)
		self._buildPanel().grid(row=1, sticky=Tkinter.NE+Tkinter.SW)
		self.statusLine = Tkinter.Label(self, relief=Tkinter.SUNKEN)
		self.statusLine.grid(row=2, sticky=Tkinter.NE+Tkinter.SW)
		self.rowconfigure(0, weight=1)
		self.columnconfigure(0, weight=1)
		self.colors = {0:"#ff0000", 1:"#00ff00", 2:"#0000ff"}
		self.pointSet = []
		self.gaussMix = []
		self._doBindings()

	def _buildPanel(self):
		butFrame = Tkinter.Frame()
		Tkinter.Button(butFrame, text="Step", command=self.runOneEmStep
			).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="Colour", 
			command=lambda: self._setColorMode(True)).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="B/W", 
			command=lambda: self._setColorMode(False)).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="Print", command=self.printMixture
			).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="50", command=lambda: 
			self.generatePointSet(50)).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="100", command=lambda: 
			self.generatePointSet(100)).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="500", command=lambda: 
			self.generatePointSet(500)).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="1000 Points", command=lambda: 
			self.generatePointSet(1000)).pack(side=Tkinter.LEFT)
		Tkinter.Button(butFrame, text="Quit", command=self.quit
			).pack(side=Tkinter.LEFT)
		return butFrame

	def _doBindings(self):
		self.bind("<space>", self.runOneEmStep)
		self.bind("q", lambda ev: self.quit())
		self.bind("p", self.printMixture)
		self.focus_set()

	def _setColorMode(self, colorMode):
		self.doColor = colorMode
		self.showPoints()

	def _enterNewGaussian(self, ev):
		newLims = self.pointDisplay.getDraggedEllipse(ev)
		try:
			self.gaussMix.addDistribution(Gaussian(Point(((newLims[0]+newLims[2])/2,
				(newLims[1]+newLims[3])/2.)), [abs(newLims[2]-newLims[0])/2.,
				abs(newLims[3]-newLims[1])/2.]))
			self.showEllipses()
		except ZeroDivisionError:
			# Rectangle too small
			pass

	def _setNewPoint(self, ev):
		self.pointSet.append(Point(self.pointDisplay.getCoos(ev)))
		self.showPoints()

	def printMixture(self, ev=None):
		print self.gaussMix

	def showPoints(self, ev=None):
		self.pointDisplay.clearPoints()
		if self.doColor:
			self.showColoredPoints()
		else:
			for p in self.pointSet:
				self.pointDisplay.setPoint(*p)

	def showColoredPoints(self, ev=None):
		for point in self.pointSet:
			probs = [g(point) for g in self.gaussMix[:3]]
			probs = probs+[0]*(3-len(probs))
			normalize(probs)
			probs = [int(255*p) for p in probs]
			self.pointDisplay.setPoint(point[0], point[1], 
				"#%02x%02x%02x"%tuple(probs))

	def _deleteGaussian(self, gaussian):
		self.gaussMix.removeDistribution(gaussian)
		self.showEllipses()

	def showEllipses(self):
		self.pointDisplay.clearEllipses()
		for index, (g, prior) in enumerate(self.gaussMix):
			center, sigmas = g.getCenter(), g.getSigmas()
			self.pointDisplay.setEllipse(center[0]-sigmas[0], center[1]-sigmas[1],
				center[0]+sigmas[0], center[1]+sigmas[1], 
				self.colors.get(index, "#000000"),
				lambda ev, gaussian=g: self._deleteGaussian(gaussian))
		self._updateStatusLine()

	def runOneEmStep(self, ev=None):
		self.gaussMix.runOneEmStep(self.pointSet)
		self.showPoints()
		self.showEllipses()

	def setMixture(self, mixture):
		self.gaussMix = mixture
		self.showEllipses()

	def generatePointSet(self, numPoints):
		self.pointSet = self.gaussMix.generatePointSet(numPoints)
		self.showPoints()
		self.showEllipses()

	def _updateStatusLine(self):
		self.statusLine.config(text="Total Log Likelihood: %d"%int(
			self.gaussMix.getTotalLogL(self.pointSet)))


def testWithGUI():
	"""runs the GUI for 2D experiments.
	"""
	g = DistMixture()
	g.addDistribution(Gaussian(Point((20,8)), [2, 4]))
	g.addDistribution(Gaussian(Point((15,16)), [4, 2]))
	g.addDistribution(Gaussian(Point((10,16)), [2, 1]))
	vis = TwoDMixtureVisualizer(600, 450, 15)
	vis.setMixture(g)
	vis.generatePointSet(100)
	vis.mainloop()


def testWithFeatureVectors():
	"""loads feature vectors and clusters them.
	"""
	import sys

	def getFeatureVectors(srcFileName):
		"""parses a feature vector file as generated by getContexts.py
		and returns the a list of feature vectors, a dictionary translating
		the index into this list to a cleartext head word and a dictionary
		translating an index in a feature vector to a cleartext word.
		"""
		# First line is the index->context word dict
		f = open(srcFileName)
		featureDict = eval(f.readline())
		headDict, featureVecs = {}, []
		# Following lines are feature vectors
		while 1:
			ln = f.readline()
			if not ln.strip():
				break
			head, vec = ln.split(":")
			headDict[len(featureVecs)] = head
			featureVecs.append(eval(vec))
		return featureVecs, headDict, featureDict

	def doEm(gaussMix):
		lastLik = gaussMix.getTotalLogL(featureVecs)
		print lastLik
		while 1:
			gaussMix.runOneEmStep(featureVecs)
			newLik = gaussMix.getTotalLogL(featureVecs)
			print newLik
			if abs(newLik-lastLik)<abs(newLik)*1e-5:
				break
			lastLik = newLik

	def output(gaussMix, featureVecs, headDict):
		clusterMembership = {}
		for index, vec in enumerate(featureVecs):
			print "%20s"%headDict[index], " ".join(
				["%10.8g"%math.log(prior*gauss(vec)) 
					for gauss, prior in gaussMix])
			clusterMembership.setdefault(gaussMix.getMostLikelyCluster(vec), []
				).append(headDict[index])
		clusters = clusterMembership.values()
		map(lambda c: c.sort(), clusters)
		clusters.sort(lambda a,b: cmp(len(a), len(b)))
		print "\n".join([" ".join(words) for words in clusters])

	def invertDict(someDict):
		return dict(starmap(lambda a, b: (b, a), someDict.iteritems()))

	def buildMixture(featureVecs, headDict):
		indexDict = invertDict(headDict)
		g = DistMixture()
		print "Known heads:", " ".join(indexDict.keys())
		defaultSigma = [0.4]*len(featureVecs[0])
		for centerTerm in sys.argv[2:]:
			try:
				g.addDistribution(Gaussian(featureVecs[indexDict[centerTerm]], 
					defaultSigma[:], minSigma=0.1))
			except KeyError:
				print "Head %s not known"%centerTerm
		return g

	featureVecs, headDict, featureDict = getFeatureVectors(sys.argv[1])
	featureVecs = [Point(v) for v in featureVecs]
	g = buildMixture(featureVecs, headDict)
	doEm(g)
	output(g, featureVecs, headDict)


if __name__=="__main__":
	testWithGUI()
#	testWithFeatureVectors()
